import axiompy

class Unit:
    def __init__(self, name, value, category, db):
        self.name = name
        self.value = value
        self.category = category
        self.db = db

    def __str__(self):
        return f"<Unit ({self.name})>"

    def __mul__(self, other):
        if(isinstance(other, (int, float, complex))):
            return axiompy.Value(other, self, self.db)
        elif(isinstance(other, axiompy.value.Value)):
            return axiompy.Value(self.value * other.value, self, self.db)

    __rmul__ = __mul__



import axiompy

class ACF: 
    """
    Axiom Configuration File, V0.1

    Sections:
        Sections are categories of variables. To create a section, start the line with a '@'.
        All ACF files require a section.
        Example:
            @length_units
            metre = 1
            centimetre = metre / 100
    Variables:
        Can be assigned values. You do not need to declare its type.
        Each variable is exclusive to it's section; it cannot be referenced from a section that it was not created in.
        You can perform math operations on the variables.
        Non-number variable values are interpreted as strings unless they are used in a math operation, where the interpreter will use it as a variable reference. 

    Math Operations:
        These operations are multiplying, dividing, adding, subtracting, and powers. (*,/,+,-,^)
        Order of operations is not yet supported (including brackets), so all operations are run executed in the order of which they were written as.
        Math operators must have a space on both sides to signify that it is an expression and not a string.
    """

    MATH_OPERATORS = {
        "ADD": "+",
        "SUBTRACT": "-",
        "MULTIPLY": "*",
        "DIVIDE": "/",
        "POWER": "^"
    }

    def __init__(self, lines):
        self.lines = lines
        self.parse()
    
    def parse(self):
        self.data = {
            "sections": {}
        }

        section = None

        for line in self.lines:
            if(line[0] == "@"): # section
                section = line[1:]
                self.data["sections"][section] = {}
            else: # section element
                if(section != None):
                    element = self.parse_element(line, section)
                    self.data["sections"][section][element["name"]] = element["value"]

    def parse_element(self, line: str, section:str):
        name = ""
        value = None

        for c in line:
            if(c in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_"):
                name += c
            elif (c == "="):
                break
        
        try:
            setter_index = line.index("=")
        except ValueError:
            ACF.throw_error(f"Setter not found in line: {line}")

        value = line[setter_index+1:].strip()

        return {"name": name, "value": self.parse_value(value, section)}

    def parse_value(self, value: str, section: str):
        tokens = self.format_tokens(value.split())

        length = len(tokens)

        i = 0
        while i < length:
            token = tokens[i]
            if(token in ACF.MATH_OPERATORS.values()):
                try:
                    answer = self.math_operation(token, tokens[i-1],tokens[i+1], section)
                    tokens.pop(i+1)
                    tokens.pop(i)
                    tokens.pop(i-1)
                    tokens.insert(i-1,answer)

                    length = len(tokens)
                    i = -1
                except KeyError:
                    ACF.throw_error(f"Math Operation without valid values - {value}")
            i += 1

        
        
        return tokens[0]

    def format_tokens(self, tokens):
        for i in range(len(tokens)):
            try:
                tokens[i] = float(tokens[i])
            except:
                pass
        return tokens

    def math_operation(self, symbol: str, value1, value2, section: str):
        ops = ACF.MATH_OPERATORS

        if(isinstance(value1, str)):
            value1 = self.get_value(value1, section)
        if(isinstance(value2, str)):
            value2 = self.get_value(value2, section)


        if(symbol == ops["MULTIPLY"]):
            return value1 * value2
        elif(symbol == ops["DIVIDE"]):
            return value1 / value2
        elif(symbol == ops["ADD"]):
            return value1 + value2
        elif(symbol == ops["SUBTRACT"]):
            return value1 - value2
        elif(symbol == ops["POWER"]):
            return value1 ** value2


    def get_value(self, name, section):
        try:
            return self.data["sections"][section][name]
        except:
            ACF.throw_error("Variable referenced before assignment. Cross-section variables are not supported.")



    @staticmethod
    def throw_error(error: str):
        raise Exception(f"Axiom Config Error: {error}")

        
    def __str__(self):
        return f"ACF<str(self.data)>"


import axiompy

class FileUtils:
    @staticmethod
    def load_acf(path):
        return axiompy.ACF(FileUtils.read_lines(path))
        
    
    @staticmethod
    def read_lines(path):
        with open(path) as f:
            return f.read().splitlines()
        


import axiompy
import math

class Value:
    
    def __init__(self, value, unit: axiompy.unit.Unit, db: axiompy.Units,dimension=1):
        self.value = value
        self.unit = unit
        self.dimension=dimension
        self.db = db

    def __str__(self):
        return f"<Value ({self.value} {self.unit}{f'^{self.dimension}' if self.dimension != 1 else ''})>"
    
    def __int__(self):
        return int(self.value)

    def incompatible_type(self, other):
        raise Exception(f"Incompatible types Value, {type(other)}")
    
    def incompatible_categories(self, other):
        raise Exception(f"Incompatible unit categories {self.unit.category}, {other.unit.category}")

    def __mul__(self, other):
        if(isinstance(other, Value)):

            if(self.unit.category != other.unit.category):
                Value.incompatible_categories(other)
            
            value1 = self.db.value_to_base(self)
            value2 = self.db.value_to_base(other)
            dimension = self.dimension + other.dimension
            base_value = Value(value1.value * value2.value, self.db.base_unit_from_value(self), dimension)
            answer = self.db.unit_convert(base_value, self.unit)

            return answer

        elif(isinstance(other, axiompy.unit.Unit)):
            if(self.unit.category != other.category):
                Value.incompatible_categories(other)

            answer = self.db.value_to_base(self) * other.value
            answer.dimension = self.dimension + 1

            return answer

        elif(isinstance(other, (int, float, complex))):
            return Value(self.value * other, self.unit, self.db)

        Value.incompatible_type(other)

    def __truediv__(self, other):
        if(isinstance(other, Value)):

            if(self.unit.category != other.unit.category):
                Value.incompatible_categories(other)
            
            value1 = self.db.value_to_base(self)
            value2 = float(self.db.value_to_base(other))
            base_value = Value(value1.value / value2.value, self.db.base_unit_from_value(self), self.dimension)
            answer = self.db.unit_convert(base_value, self.unit)

            return answer

        elif(isinstance(other, axiompy.unit.Unit)):
            if(self.unit.category != other.category):
                Value.incompatible_categories(other)

            answer = self.db.value_to_base(self) / float(other.value)

            return answer

        elif(isinstance(other, (int, float, complex))):
            return Value(self.value / other, self.unit, self.db)

        Value.incompatible_type(other)

    def __floordiv__(self, other):
        if(isinstance(other, Value)):

            if(self.unit.category != other.unit.category):
                Value.incompatible_categories(other)
            
            value1 = self.db.value_to_base(self)
            value2 = float(self.db.value_to_base(other))
            base_value = Value(math.floor(value1.value / value2.value), self.db.base_unit_from_value(self), self.dimension)
            answer = self.db.unit_convert(base_value, self.unit)

            return answer

        elif(isinstance(other, axiompy.unit.Unit)):
            if(self.unit.category != other.category):
                Value.incompatible_categories(other)

            answer = math.floor(self.db.value_to_base(self) / float(other.value))

            return answer

        elif(isinstance(other, (int, float, complex))):
            return Value(math.floor(self.value / other), self.unit, self.db)

        Value.incompatible_type(other)


    def __add__(self, other):
        if(isinstance(other, Value)):

            if(self.unit.category != other.unit.category):
                Value.incompatible_categories(other)
            
            value1 = self.db.value_to_base(self)
            value2 = self.db.value_to_base(other)
            base_value = Value(value1.value + value2.value, self.db.base_unit_from_value(self), self.dimension)
            answer = self.db.unit_convert(base_value, self.unit)

            return answer

        elif(isinstance(other, axiompy.unit.Unit)):
            if(self.unit.category != other.category):
                Value.incompatible_categories(other)

            answer = self.db.value_to_base(self) + other.value

            return answer

        elif(isinstance(other, (int, float, complex))):
            return Value(self.value + other, self.unit, self.db)

        Value.incompatible_type(other)
    
    def __sub__(self, other):
        if(isinstance(other, Value)):

            if(self.unit.category != other.unit.category):
                Value.incompatible_categories(other)
            
            value1 = self.db.value_to_base(self)
            value2 = self.db.value_to_base(other)
            base_value = Value(value1.value - value2.value, self.db.base_unit_from_value(self), self.dimension)
            answer = self.db.unit_convert(base_value, self.unit)

            return answer

        elif(isinstance(other, axiompy.unit.Unit)):
            if(self.unit.category != other.category):
                Value.incompatible_categories(other)

            answer = self.db.value_to_base(self) - other.value

            return answer

        elif(isinstance(other, (int, float, complex))):
            return Value(self.value - other, self.unit, self.db)

        Value.incompatible_type(other)

    __rmul__ = __mul__
    __rtruediv__ = __truediv__
    __rfloordiv__ = __floordiv__
    __radd__ = __add__
    __rsub__ = __sub__


import axiompy
import pkg_resources

class Units:
    def __init__(self):

        rsman = pkg_resources.ResourceManager()

        unit_file = rsman.resource_string('axiompy', 'db/en_units.acf').decode('utf-8')
        lines = unit_file.splitlines()

        lines = list(filter(None, lines)) #Remove empty lines

        self.units_dict = axiompy.ACF(lines).data

        self.units = self.acf_dict_to_units(self.units_dict)

    def acf_dict_to_units(self, d):
        """
        Converts the dictionary generated by the ACF to units
        so that the unit doesn't need to be found every time it is called for - to improve performance.
        """

        units = {"base_units": {}}

        for category in d["sections"].keys():
            for unit in d["sections"][category].keys():
                if(unit == "base_unit"):
                    units["base_units"][category] = d["sections"][category]["base_unit"]
                else:
                    value = d["sections"][category][unit]
                    units[unit] = axiompy.Unit(unit, value, category, self)

        return units

    def unit_convert(self, from_value, to_unit):
        """
        Params:
            from: axiompy.Value
            tounit: axiompy.Unit

        Converts an axiompy value to a desired unit.
        """

        from_unit = from_value.unit
        
        if(from_unit.category != to_unit.category):
            raise Exception(f"Incompatible unit categories {from_unit.category}, {to_unit.category}")

        to_unit_value = to_unit.value

        return axiompy.Value((from_value.value * from_unit.value) / to_unit_value, to_unit, self, dimension=from_value.dimension)

    def base_unit_from_string(self, unit_cat: str):
        return self.units[self.units["base_units"][unit_cat]]

    def base_unit_from_value(self, value):
        return self.base_unit_from_string(value.unit.category)

    def value_to_base(self, value):
        """
        Converts a value's unit to its base unit
        ie 15cm converts to 0.15m as metre is the base unit for length.

        >>> unit_to_base(Value(15, AxiomPy.centimetre))
        Value(0.15, AxiomPy.metre)
        """

        base_unit = self.base_unit_from_value(value)

        return axiompy.Value(value.unit.value * value.value, base_unit, self)        

    def __getattr__(self, name):
        value = self.units[name]

        return value


from setuptools import find_packages
from distutils.core import setup

setup(name='axiompy',
      version='0.1.3',
      description='Unit Conversions and other math',
      author='ArztKlein',
      author_email='',
      url='https://github.com/ArztKlein/Axiompy',
      keywords=['python','unit conversion', 'units'],
      setup_requires=['wheel'],
      python_requires='>3.6.1',
      packages=find_packages(),
      include_package_data=True
     )